from __future__ import annotations

from typing import Any, List, cast

from pyasdl.grammar import AST


class Visitor:
    def __call__(self, node: AST, *args, **kwargs) -> Any:
        ...


class ASDLVisitor:
    def visit(self, node: AST, *args, **kwargs) -> Any:
        visitor = self.find_visitor(type(node).__name__)
        return visitor(node, *args, **kwargs)

    def visit_all(self, nodes: List[AST], *args, **kwags) -> List[Any]:
        return [self.visit(node, *args, **kwags) for node in nodes]

    def generic_visit(self, node: AST, *args, **kwags) -> AST:
        for value in vars(node).values():
            if isinstance(value, AST):
                self.visit(value)
            elif isinstance(value, list):
                self.visit_all(value)
        return node

    def find_visitor(self, name: str) -> Visitor:
        visitor = f"visit_{name}"
        if hasattr(self, visitor):
            func = getattr(self, visitor)
        else:
            func = self.generic_visit
        return cast(Visitor, func)



#!/usr/bin/env python3.8
# @generated by pegen from pyasdl/static/asdl.gram


from typing import Any, List, Optional, Union

from pegen.parser import Parser, memoize

from pyasdl.grammar import (
    Constructor,
    Field,
    FieldQualifier,
    Module,
    Product,
    Sum,
    Type,
)

TypeList = List[Type]
FieldList = List[Field]
ConstructorList = List[Constructor]
ProductOrSum = Union[Product, Sum]

# Keywords and soft keywords are listed at the end of the parser definition.
class GeneratedParser(Parser):
    @memoize
    def start(self) -> Optional[Module]:
        # start: "module" NAME NEWLINE? "{" definitions "}"
        mark = self._mark()
        if (
            (literal := self.expect("module"))
            and (name := self.name())
            and (opt := self.expect("NEWLINE"),)
            and (literal_1 := self.expect("{"))
            and (body := self.definitions())
            and (literal_2 := self.expect("}"))
        ):
            return Module(name.string, body)
        self._reset(mark)
        return None

    @memoize
    def definitions(self) -> Optional[TypeList]:
        # definitions: definition definitions | definition
        mark = self._mark()
        if (definition := self.definition()) and (
            definitions := self.definitions()
        ):
            return [definition] + definitions
        self._reset(mark)
        if definition := self.definition():
            return [definition]
        self._reset(mark)
        return None

    @memoize
    def definition(self) -> Optional[Type]:
        # definition: NAME "=" define
        mark = self._mark()
        if (
            (name := self.name())
            and (literal := self.expect("="))
            and (define := self.define())
        ):
            return Type(name.string, define)
        self._reset(mark)
        return None

    @memoize
    def define(self) -> Optional[ProductOrSum]:
        # define: sum_body attributes? | fields attributes?
        mark = self._mark()
        if (sum_body := self.sum_body()) and (attrs := self.attributes(),):
            return Sum(sum_body, attrs or [])
        self._reset(mark)
        if (fields := self.fields()) and (attrs := self.attributes(),):
            return Product(fields, attrs or [])
        self._reset(mark)
        return None

    @memoize
    def sum_body(self) -> Optional[ConstructorList]:
        # sum_body: constructor !"|" | "|".constructor+
        mark = self._mark()
        if (constructor := self.constructor()) and self.negative_lookahead(
            self.expect, "|"
        ):
            return [constructor]
        self._reset(mark)
        if constructors := self._gather_1():
            return constructors
        self._reset(mark)
        return None

    @memoize
    def constructor(self) -> Optional[Constructor]:
        # constructor: NAME fields?
        mark = self._mark()
        if (name := self.name()) and (fields := self.fields(),):
            return Constructor(name.string, fields or [])
        self._reset(mark)
        return None

    @memoize
    def fields(self) -> Optional[FieldList]:
        # fields: "(" fields_body? ")"
        mark = self._mark()
        if (
            (literal := self.expect("("))
            and (body := self.fields_body(),)
            and (literal_1 := self.expect(")"))
        ):
            return body
        self._reset(mark)
        return None

    @memoize
    def fields_body(self) -> Optional[FieldList]:
        # fields_body: field !"," | ",".field+
        mark = self._mark()
        if (field := self.field()) and self.negative_lookahead(
            self.expect, ","
        ):
            return [field]
        self._reset(mark)
        if fields := self._gather_3():
            return fields
        self._reset(mark)
        return None

    @memoize
    def field(self) -> Optional[Field]:
        # field: NAME field_qualifier? NAME
        mark = self._mark()
        if (
            (type := self.name())
            and (qualifier := self.field_qualifier(),)
            and (name := self.name())
        ):
            return Field(type.string, name.string, qualifier)
        self._reset(mark)
        return None

    @memoize
    def field_qualifier(self) -> Optional[FieldQualifier]:
        # field_qualifier: "?" | "*"
        mark = self._mark()
        if literal := self.expect("?"):
            return FieldQualifier.OPTIONAL
        self._reset(mark)
        if literal := self.expect("*"):
            return FieldQualifier.SEQUENCE
        self._reset(mark)
        return None

    @memoize
    def attributes(self) -> Optional[FieldList]:
        # attributes: "attributes" fields
        mark = self._mark()
        if (literal := self.expect("attributes")) and (
            fields := self.fields()
        ):
            return fields
        self._reset(mark)
        return None

    @memoize
    def _loop0_2(self) -> Optional[Any]:
        # _loop0_2: "|" constructor
        mark = self._mark()
        children = []
        while (literal := self.expect("|")) and (elem := self.constructor()):
            children.append(elem)
            mark = self._mark()
        self._reset(mark)
        return children

    @memoize
    def _gather_1(self) -> Optional[Any]:
        # _gather_1: constructor _loop0_2
        mark = self._mark()
        if (elem := self.constructor()) is not None and (
            seq := self._loop0_2()
        ) is not None:
            return [elem] + seq
        self._reset(mark)
        return None

    @memoize
    def _loop0_4(self) -> Optional[Any]:
        # _loop0_4: "," field
        mark = self._mark()
        children = []
        while (literal := self.expect(",")) and (elem := self.field()):
            children.append(elem)
            mark = self._mark()
        self._reset(mark)
        return children

    @memoize
    def _gather_3(self) -> Optional[Any]:
        # _gather_3: field _loop0_4
        mark = self._mark()
        if (elem := self.field()) is not None and (
            seq := self._loop0_4()
        ) is not None:
            return [elem] + seq
        self._reset(mark)
        return None

    KEYWORDS = ()
    SOFT_KEYWORDS = ("attributes", "module")


if __name__ == "__main__":
    from pegen.parser import simple_parser_main

    simple_parser_main(GeneratedParser)



import argparse
import io
import tokenize as _tokenize
from typing import Iterator

from pegen.tokenizer import Tokenizer

from pyasdl.grammar import Module, Sum
from pyasdl.parser import GeneratedParser as _ASDLParser

__all__ = ["parse", "fetch_comments", "is_simple_sum"]
# Since the pegen.tokenizer.Tokenizer uses .type instead of .exact_type
# it is not trivial to change the default comment behavior. A workaround
# way is sanitizing the input before passing it into the real parser

COMMENT_PATTERN = _tokenize.Whitespace + r"--.*?\n"
_tokenize.PseudoToken = _tokenize.Whitespace + _tokenize.group(
    COMMENT_PATTERN,
    _tokenize.PseudoExtras,
    _tokenize.Number,
    _tokenize.Funny,
    _tokenize.ContStr,
    _tokenize.Name,
)


def tokenize(
    source: str, ignore_comments: bool = True
) -> Iterator[_tokenize.TokenInfo]:
    # A wrapper around tokenize.generate_tokens to pass comment tokens
    source_buffer = io.StringIO(source)
    for token in _tokenize.generate_tokens(source_buffer.readline):
        if token.string.startswith("--") and ignore_comments:
            continue
        yield token


def parse(source: str, *, filename: str = "<pyasdl>") -> Module:
    """Parse the given `source` string, and return
    the AST in the shape of an `pyasdl.Module`."""

    tokenizer = Tokenizer(tokenize(source))
    parser = _ASDLParser(tokenizer)
    tree = parser.start()

    if tree is None:
        raise parser.make_syntax_error(filename)
    return tree


def fetch_comments(source: str) -> Iterator[str]:
    """Return an iterator of the ASDL comments in the
    given `source` string."""
    for token in tokenize(source, ignore_comments=False):
        if token.string.startswith("--"):
            yield token.string[2:]


def is_simple_sum(node: Sum) -> bool:
    """Check whether if the given `node`'s
    all children lack any fields."""
    return (
        all(len(constructor.fields) == 0 for constructor in node.types)
        and len(node.attributes) == 0
    )


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    options = parser.parse_args()

    with open(options.file) as f:
        content = f.read()

    print(parse(content))


if __name__ == "__main__":
    main()



# @generated by generators/python.py from pyasdl/static/grammar.asdl

from __future__ import annotations

import typing
from dataclasses import dataclass as _dataclass
from dataclasses import field as _field
from enum import Enum as _Enum
from enum import auto as _auto

identifier = str
string = typing.Union[
    str, bytes
]  # Can't use AnyStr before PEP 613 is supported
constant = typing.Union[  # type: ignore
    str,
    bytes,
    # strings
    int,
    float,
    complex,
    # numbers
    bool,
    # other
    tuple,
    frozenset,
    # sequences
    None,
    type(Ellipsis)  # type: ignore
    # singletons
]


class AST:
    ...


@_dataclass
class Module(AST):
    name: string
    body: typing.List[Type] = _field(default_factory=list)


@_dataclass
class Type(AST):
    name: string
    value: type


class type(AST):
    ...


@_dataclass
class Sum(type):
    types: typing.List[Constructor] = _field(default_factory=list)
    attributes: typing.List[Field] = _field(default_factory=list)


@_dataclass
class Product(type):
    fields: typing.List[Field] = _field(default_factory=list)
    attributes: typing.List[Field] = _field(default_factory=list)


@_dataclass
class Constructor(AST):
    name: string
    fields: typing.List[Field] = _field(default_factory=list)


@_dataclass
class Field(AST):
    kind: string
    name: string
    qualifier: typing.Optional[FieldQualifier] = _field(default=None)


class FieldQualifier(_Enum):
    OPTIONAL = _auto()
    SEQUENCE = _auto()
